""" Interface with `ROS <https://www.ros.org/>`_.

Functions
---------

.. autosummary::
    :toctree: _autosummary

    create_frame_graph

"""

from .generic import create_frame_graph

__all__ = ["create_frame_graph"]
